"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ActionHandlersRegistry = void 0;
const FBLService_1 = require("./FBLService");
const LogService_1 = require("./LogService");
class ActionHandlersRegistry {
    constructor() {
        this.cleanup();
    }
    static get instance() {
        if (!this.pInstance) {
            this.pInstance = new ActionHandlersRegistry();
        }
        return this.pInstance;
    }
    static reset() {
        this.pInstance = null;
    }
    /**
     * Register new action handler
     * @param {ActionHandler} handler
     * @param {IPlugin} plugin
     * @param {ActionSnapshot} [snapshot]
     * @return self instance
     */
    register(handler, plugin, snapshot) {
        if (handler.getMetadata().aliases) {
            for (const alias of handler.getMetadata().aliases) {
                if (alias.startsWith(FBLService_1.FBLService.METADATA_PREFIX)) {
                    throw new Error(`Unable to register action handler with alias "${alias}". It can not start with ${FBLService_1.FBLService.METADATA_PREFIX}`);
                }
            }
        }
        if (handler.getMetadata().id.startsWith(FBLService_1.FBLService.METADATA_PREFIX)) {
            throw new Error(`Unable to register action handler with id "${handler.getMetadata().id}". It can not start with ${FBLService_1.FBLService.METADATA_PREFIX}`);
        }
        const metadata = handler.getMetadata();
        if (this.registry[metadata.id]) {
            if (snapshot) {
                snapshot.log(`Action handler with id ${metadata.id} was overridden by plugin ${plugin.name}@${plugin.version}`, true);
            }
            else {
                LogService_1.LogService.instance.error(' -> Warning'.red +
                    ' action handler with id: ' +
                    metadata.id.red +
                    ' was overridden by plugin ' +
                    `${plugin.name}@${plugin.version}`.red);
            }
        }
        this.registry[metadata.id] = handler;
        if (metadata.aliases) {
            metadata.aliases.forEach((alias) => {
                if (this.aliases[alias]) {
                    /* istanbul ignore else */
                    if (snapshot) {
                        snapshot.log(`Action handler with alias ${alias} was overridden by plugin ${plugin.name}@${plugin.version}`, true);
                    }
                    else {
                        LogService_1.LogService.instance.error(' -> Warning'.red +
                            ' action handler with alias: ' +
                            alias.red +
                            ' was overridden by plugin ' +
                            `${plugin.name}@${plugin.version}`.red);
                    }
                }
                this.aliases[alias] = metadata.id;
            });
        }
        return this;
    }
    /**
     * Unregister handler by name
     * @param {string} id
     * @return self instance
     */
    unregister(id) {
        delete this.registry[id];
        Object.keys(this.aliases).forEach((key) => {
            if (this.aliases[key] === id) {
                delete this.aliases[key];
            }
        });
        return this;
    }
    cleanup() {
        this.registry = {};
        this.aliases = {};
        return this;
    }
    /**
     * Find action handler by id or alias
     * @param {string} idOrAlias
     */
    find(idOrAlias) {
        let actionHandler = this.registry[idOrAlias];
        if (!actionHandler) {
            const id = this.aliases[idOrAlias];
            if (id) {
                actionHandler = this.registry[id];
            }
        }
        if (!actionHandler) {
            return null;
        }
        return actionHandler;
    }
}
exports.ActionHandlersRegistry = ActionHandlersRegistry;
//# sourceMappingURL=ActionHandlersRegistry.js.map