"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FlowService = void 0;
const ActionHandlersRegistry_1 = require("./ActionHandlersRegistry");
const js_yaml_1 = require("js-yaml");
const ejs = require("ejs");
const models_1 = require("../models");
const utils_1 = require("../utils");
const TemplateUtilitiesRegistry_1 = require("./TemplateUtilitiesRegistry");
const path_1 = require("path");
const tar_1 = require("tar");
const fs_1 = require("fs");
const util_1 = require("util");
const got_1 = require("got");
const TempPathsRegistry_1 = require("./TempPathsRegistry");
const os_1 = require("os");
const LogService_1 = require("./LogService");
const object_collider_1 = require("object-collider");
const FBLService_1 = require("./FBLService");
const uuid_1 = require("uuid");
const ejsLint = require('ejs-lint');
class FlowService {
    constructor() {
        this.index = 0;
        this.flowPathCache = {};
        this.flowResolvers = {};
        /**
         * Turn to true to capture snapshots
         * @type {boolean}
         */
        this.debug = false;
    }
    static get instance() {
        if (!this.pInstance) {
            this.pInstance = new FlowService();
        }
        return this.pInstance;
    }
    static reset() {
        this.pInstance = null;
    }
    /**
     * Get FBL home directory
     */
    static getHomeDir() {
        return process.env.FBL_HOME || path_1.resolve(os_1.homedir(), '.fbl');
    }
    /**
     * Execute action
     * @param {string} wd Working Directory
     * @param action
     * @param {IContext} context
     * @param {IDelegatedParameters} parameters
     * @param {ActionSnapshot} parentSnapshot
     * @returns {Promise<void>}
     */
    executeAction(source, wd, action, context, parameters, parentSnapshot) {
        return __awaiter(this, void 0, void 0, function* () {
            let options;
            let idOrAlias;
            let metadata;
            if (typeof action !== 'string') {
                if (!parentSnapshot) {
                    parentSnapshot = new models_1.ActionSnapshot(source, 'void', {}, wd, 0, {});
                }
                idOrAlias = FBLService_1.FBLService.extractIdOrAlias(action);
                options = action[idOrAlias];
                metadata = FBLService_1.FBLService.extractMetadata(action);
                metadata = yield this.resolveOptionsWithNoHandlerCheck(context.ejsTemplateDelimiters.local, metadata, context, parentSnapshot, parameters, false);
            }
            else {
                idOrAlias = action;
                metadata = {};
            }
            if (metadata && metadata.$parameters) {
                utils_1.ContextUtil.assign(parameters.parameters, '$', metadata.$parameters, false);
            }
            const idx = ++this.index;
            LogService_1.LogService.instance.info(` -> [${idx}] [${source}] [${(metadata && metadata.$title) || idOrAlias}]`.green + ' Processing.');
            let snapshot;
            if (this.debug) {
                snapshot = new models_1.EnabledActionSnapshot(source, idOrAlias, metadata, wd, idx, parameters);
            }
            else {
                snapshot = new models_1.ActionSnapshot(source, idOrAlias, metadata, wd, idx, parameters);
            }
            try {
                let handler = ActionHandlersRegistry_1.ActionHandlersRegistry.instance.find(idOrAlias);
                if (!handler) {
                    handler = context.dynamicActionHandlers.find(idOrAlias);
                }
                if (!handler) {
                    throw new Error(`Unable to find action handler for: ${idOrAlias}`);
                }
                const alternativeWD = handler.getWorkingDirectory();
                if (alternativeWD) {
                    snapshot.wd = alternativeWD;
                }
                snapshot.setInitialContextState(context);
                snapshot.setActionHandlerId(handler.getMetadata().id);
                if (!handler.getMetadata().considerOptionsAsSecrets) {
                    snapshot.setOptions(options);
                    // register options twice to see what's actually has been changed (only when changes applied)
                    const resolvedOptions = yield this.resolveOptions(handler, options, context, snapshot, parameters, true);
                    if (JSON.stringify(options) !== JSON.stringify(resolvedOptions)) {
                        snapshot.setOptions(resolvedOptions);
                    }
                }
                else {
                    snapshot.setOptions(FlowService.MASKED);
                }
                // resolve without masking
                options = yield this.resolveOptions(handler, options, context, snapshot, parameters, false);
                const processor = handler.getProcessor(options, context, snapshot, parameters);
                yield processor.validate();
                snapshot.validated();
                const shouldExecute = yield processor.isShouldExecute();
                if (shouldExecute) {
                    snapshot.start();
                    yield processor.execute();
                    snapshot.success();
                    if (snapshot.successful) {
                        LogService_1.LogService.instance.info(` <- [${idx}] [${source}] [${(metadata && metadata.$title) || idOrAlias}]`.blue +
                            ' Completed successfully withing ' +
                            snapshot.getHumanReadableDuration().blue);
                    }
                    else {
                        LogService_1.LogService.instance.info(` <- [${idx}] [${source}] [${(metadata && metadata.$title) || idOrAlias}]`.yellow +
                            ' Marked as failed. Took ' +
                            snapshot.getHumanReadableDuration().yellow);
                    }
                }
                else {
                    LogService_1.LogService.instance.info(` <- [${idx}] [${source}] [${(metadata && metadata.$title) || idOrAlias}]`.yellow + ' Skipped');
                    snapshot.skipped();
                }
            }
            catch (e) {
                LogService_1.LogService.instance.error(` <- [${idx}] [${source}] [${(metadata && metadata.$title) || idOrAlias}]`.red +
                    ` Failed with: ${e.toString().red}`);
                snapshot.failure(e);
            }
            return snapshot;
        });
    }
    /**
     * Get cached tarball pass by given url
     * @param url
     */
    static getCachedTarballPathForURL(url) {
        return __awaiter(this, void 0, void 0, function* () {
            const cacheDir = path_1.resolve(FlowService.getHomeDir(), 'cache');
            yield utils_1.FSUtil.mkdirp(cacheDir);
            return path_1.resolve(cacheDir, uuid_1.v5(url, uuid_1.v5.URL) + '.tar.gz');
        });
    }
    /**
     * Download tarball into temp location
     * @param {IFlowLocationOptions} location
     * @param {number} [redirectCount]
     * @return {Promise<string>} temp tarball location
     */
    downloadTarball(location, redirectCount = 0) {
        return __awaiter(this, void 0, void 0, function* () {
            let tarballFile;
            if (location.cache) {
                tarballFile = yield FlowService.getCachedTarballPathForURL(location.path);
                const exists = yield utils_1.FSUtil.exists(tarballFile);
                if (exists) {
                    LogService_1.LogService.instance.info(' -> Found cached tarball for remote URL: '.green + location.path);
                    return tarballFile;
                }
            }
            else {
                tarballFile = yield TempPathsRegistry_1.TempPathsRegistry.instance.createTempFile(false, '.tar.gz');
            }
            LogService_1.LogService.instance.info(' -> Downloading tarball from remote URL: '.green + location.path + ' to ' + tarballFile);
            const ws = fs_1.createWriteStream(tarballFile);
            try {
                yield new Promise((res, rej) => {
                    const stream = got_1.default.stream(location.path, {
                        isStream: true,
                        timeout: 120 * 1000,
                        headers: location.http && location.http.headers,
                    });
                    stream.pipe(ws);
                    ws.on('finish', res);
                    stream.on('error', rej);
                });
            }
            catch (e) {
                const exists = yield utils_1.FSUtil.exists(tarballFile);
                /* istanbul ignore else */
                if (exists) {
                    yield util_1.promisify(fs_1.unlink)(tarballFile);
                }
                throw e;
            }
            return tarballFile;
        });
    }
    /**
     * Extract tarball to temp dir
     * @param {string} path
     * @return {Promise<string>} path to temp dir
     */
    extractTarball(path) {
        return __awaiter(this, void 0, void 0, function* () {
            LogService_1.LogService.instance.info(' -> Extracting tarball at path: '.green + path);
            const tarball = path;
            const result = yield TempPathsRegistry_1.TempPathsRegistry.instance.createTempDir();
            yield tar_1.x({
                file: tarball,
                C: result,
            });
            return result;
        });
    }
    /**
     * Recursively find index.yml in directory structure
     * @param {string} path
     * @return {Promise<string>}
     */
    recursivelyFindIndexFileInDir(path) {
        return __awaiter(this, void 0, void 0, function* () {
            let contents = yield util_1.promisify(fs_1.readdir)(path);
            // filter all files and folders that start with "."
            contents = contents.filter((fileOrDir) => !fileOrDir.startsWith('.'));
            const match = contents.find((dirOrFile) => {
                return dirOrFile === 'index.yml' || dirOrFile === 'index.yaml';
            });
            if (match) {
                const dirOrFile = path_1.join(path, match);
                const isDir = yield utils_1.FSUtil.isDirectory(dirOrFile);
                if (!isDir) {
                    return dirOrFile;
                }
            }
            if (contents.length === 1) {
                const dirOrFile = path_1.join(path, contents[0]);
                const isDir = yield utils_1.FSUtil.isDirectory(dirOrFile);
                if (isDir) {
                    return yield this.recursivelyFindIndexFileInDir(dirOrFile);
                }
            }
            throw new Error('Unable to locate index file inside the directory.');
        });
    }
    /**
     * Resolve flow, skipping checks if similar resolve action is already running
     * @param {IFlowLocationOptions} location
     * @return {Promise<string>}
     */
    resolveFlowSkipChecks(location) {
        return __awaiter(this, void 0, void 0, function* () {
            let absolutePath = location.path;
            if (location.path.startsWith('http://') || location.path.startsWith('https://')) {
                absolutePath = yield this.downloadTarball(location);
            }
            // if path leads to tarball - extract it to temp dir
            if (absolutePath.endsWith('.tar.gz')) {
                absolutePath = yield this.extractTarball(absolutePath);
            }
            this.flowPathCache[location.path] = absolutePath;
            return absolutePath;
        });
    }
    resolveFlowTarget(local, location) {
        return __awaiter(this, void 0, void 0, function* () {
            // if path lead to directory - use index.yml inside it as a starting point
            const directory = yield utils_1.FSUtil.isDirectory(local);
            if (directory) {
                if (location.target) {
                    local = path_1.join(local, location.target);
                }
                else {
                    local = yield this.recursivelyFindIndexFileInDir(local);
                }
            }
            else {
                if (location.target) {
                    throw new Error(`Usage of target is not allowed for flow at path: ${location.path}`);
                }
            }
            return local;
        });
    }
    /**
     * Resolve flow based on path
     * @param {IFlowLocationOptions} location
     * @return {Promise<string>}
     */
    resolveFlow(location) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!utils_1.FSUtil.isAbsolute(location.path) && !utils_1.FSUtil.isURL(location.path)) {
                throw new Error(`Provided path ${location.path} is not absolute.`);
            }
            if (this.flowPathCache[location.path]) {
                const cachedPath = this.flowPathCache[location.path];
                return yield this.resolveFlowTarget(cachedPath, location);
            }
            let resolver = this.flowResolvers[location.path];
            if (resolver) {
                const cachedPath = yield resolver;
                return yield this.resolveFlowTarget(cachedPath, location);
            }
            resolver = this.resolveFlowSkipChecks(location);
            this.flowResolvers[location.path] = resolver;
            const result = yield resolver;
            delete this.flowResolvers[location.path];
            return yield this.resolveFlowTarget(result, location);
        });
    }
    /**
     * Read flow from file
     * @param {IFlowLocationOptions} location
     * @param {IContext} context
     * @param {ActionSnapshot} snapshot
     * @param {IDelegatedParameters} parameters
     * @returns {Promise<IFlow>}
     */
    readFlowFromFile(location, context, snapshot, parameters) {
        return __awaiter(this, void 0, void 0, function* () {
            const absolutePath = yield this.resolveFlow(location);
            LogService_1.LogService.instance.info(` -> Reading flow file: `.green + absolutePath);
            let content = yield utils_1.FSUtil.readTextFile(absolutePath);
            content = yield this.resolveTemplate(context.ejsTemplateDelimiters.global, content, context, snapshot, parameters);
            let flow;
            try {
                flow = js_yaml_1.load(content);
            }
            catch (e) {
                LogService_1.LogService.instance.error(` -> Reading flow failed from file: `.red + absolutePath + ' Error: ' + e.message.red);
                LogService_1.LogService.instance.error(content.gray);
                throw e;
            }
            return { flow, wd: path_1.dirname(absolutePath) };
        });
    }
    /**
     * Resolve template
     * @param {string} delimiter EJS template delimiter
     * @param {string} tpl template to resolve
     * @param {IContext} context
     * @param {ActionSnapshot} snapshot
     * @param {IDelegatedParameters} parameters
     * @param [extraVariables] additional varialbes to be presented in EJS data
     * @return {string}
     */
    resolveTemplate(delimiter, tpl, context, snapshot, parameters, extraVariables, wd) {
        return __awaiter(this, void 0, void 0, function* () {
            // validate template
            ejsLint(tpl, { delimiter });
            const data = Object.assign(Object.assign({ $: TemplateUtilitiesRegistry_1.TemplateUtilitiesRegistry.instance.generateUtilities(context, snapshot, parameters, wd || snapshot.wd), env: process.env }, parameters), context);
            if (extraVariables) {
                Object.assign(data, extraVariables);
            }
            const options = {
                delimiter,
                escape: (value) => {
                    if (typeof value === 'number' || typeof value === 'boolean') {
                        return value.toString();
                    }
                    if (typeof value === 'string') {
                        return JSON.stringify(value);
                    }
                    if (object_collider_1.isMissing(value)) {
                        return 'null';
                    }
                    throw Error(`Value could not be escaped. Use $ref:path to pass value by reference.`);
                },
                async: true,
            };
            const result = yield ejs.render(tpl, data, options);
            return result;
        });
    }
    /**
     * Resolve options with no handler check
     * @param {string} delimiter - EJS template delimiter, by default EJS is using %
     * @param options
     * @param {IContext} context
     * @param {ActionSnapshot} snapshot
     * @param {IDelegatedParameters} parameters delegated parameters
     * @param {boolean} maskSecrets if true - all secrets will be masked
     * @return {any}
     */
    resolveOptionsWithNoHandlerCheck(delimiter, options, context, snapshot, parameters, maskSecrets) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!options) {
                return options;
            }
            if (maskSecrets && context.secrets && Object.keys(context.secrets).length) {
                // make a copy of the options object first
                let json = JSON.stringify(options);
                const pattern = `<${delimiter}[^${delimiter}>|\\w]*[^\\w]+(secrets[^\\w])[^${delimiter}>]*${delimiter}>`;
                const regex = new RegExp(pattern, 'g');
                json = json.replace(regex, FlowService.MASKED);
                options = JSON.parse(json);
            }
            const tpl = js_yaml_1.dump(options);
            const yaml = yield this.resolveTemplate(delimiter, tpl, context, snapshot, parameters);
            options = js_yaml_1.load(yaml);
            // resolve references
            options = utils_1.ContextUtil.resolveReferences(options, context, parameters);
            return options;
        });
    }
    /**
     * Resolve options for handler
     * @param {ActionHandler} handler
     * @param options
     * @param {IContext} context
     * @param {IDelegatedParameters} parameters delegated parameters
     * @param {boolean} maskSecrets if true - all secrets will be masked
     * @returns {Promise<any>}
     */
    resolveOptions(handler, options, context, snapshot, parameters, maskSecrets) {
        return __awaiter(this, void 0, void 0, function* () {
            if (handler.getMetadata().skipTemplateProcessing) {
                return options;
            }
            return yield this.resolveOptionsWithNoHandlerCheck(context.ejsTemplateDelimiters.local, options, context, snapshot, parameters, maskSecrets);
        });
    }
}
exports.FlowService = FlowService;
FlowService.MASKED = '{MASKED}';
//# sourceMappingURL=FlowService.js.map